LIBRARY  isa-l_crypto
EXPORTS

sha1_ctx_mgr_init                      @1
sha1_ctx_mgr_submit                    @2
sha1_ctx_mgr_flush                     @3
sha256_ctx_mgr_init                    @4
sha256_ctx_mgr_submit                  @5
sha256_ctx_mgr_flush                   @6
sha512_ctx_mgr_init                    @7
sha512_ctx_mgr_submit                  @8
sha512_ctx_mgr_flush                   @9
md5_ctx_mgr_init                       @10
md5_ctx_mgr_submit                     @11
md5_ctx_mgr_flush                      @12
mh_sha1_init                           @13
mh_sha1_update                         @14
mh_sha1_finalize                       @15
mh_sha1_finalize_base                  @16
mh_sha1_update_base                    @17
mh_sha1_murmur3_x64_128_init           @18
mh_sha1_murmur3_x64_128_finalize_base  @19
mh_sha1_murmur3_x64_128_update_base    @20
mh_sha1_murmur3_x64_128_update         @21
mh_sha1_murmur3_x64_128_finalize       @22
aes_keyexp_128                         @23
aes_keyexp_192                         @24
aes_keyexp_256                         @25
aesni_gcm128_enc                       @26
aesni_gcm128_dec                       @27
aesni_gcm128_init                      @28
aesni_gcm128_enc_update                @29
aesni_gcm128_dec_update                @30
aesni_gcm128_enc_finalize              @31
aesni_gcm128_dec_finalize              @32
aesni_gcm256_enc                       @33
aesni_gcm256_dec                       @34
aesni_gcm256_init                      @35
aesni_gcm256_enc_update                @36
aesni_gcm256_dec_update                @37
aesni_gcm256_enc_finalize              @38
aesni_gcm256_dec_finalize              @39
aesni_gcm256_pre                       @40
aesni_gcm128_pre                       @41
aes_cbc_enc_128                        @42
aes_cbc_dec_128                        @43
aes_cbc_enc_192                        @44
aes_cbc_dec_192                        @45
aes_cbc_enc_256                        @46
aes_cbc_dec_256                        @47
aes_cbc_precomp                        @48
XTS_AES_128_enc                        @49
XTS_AES_128_enc_expanded_key           @50
XTS_AES_128_dec                        @51
XTS_AES_128_dec_expanded_key           @52
XTS_AES_256_enc                        @53
XTS_AES_256_enc_expanded_key           @54
XTS_AES_256_dec                        @55
XTS_AES_256_dec_expanded_key           @56
mh_sha256_init                         @57
mh_sha256_update                       @58
mh_sha256_finalize                     @59
mh_sha256_finalize_base                @60
mh_sha256_update_base                  @61
rolling_hashx_mask_gen                 @62
rolling_hash2_run                      @63
rolling_hash2_reset                    @64
rolling_hash2_init                     @65
aes_gcm_pre_128                        @66
aes_gcm_enc_128                        @67
aes_gcm_dec_128                        @68
aes_gcm_init_128                       @69
aes_gcm_enc_128_update                 @70
aes_gcm_dec_128_update                 @71
aes_gcm_enc_128_finalize               @72
aes_gcm_dec_128_finalize               @73
aes_gcm_pre_256                        @74
aes_gcm_enc_256                        @75
aes_gcm_dec_256                        @76
aes_gcm_init_256                       @77
aes_gcm_enc_256_update                 @78
aes_gcm_dec_256_update                 @79
aes_gcm_enc_256_finalize               @80
aes_gcm_dec_256_finalize               @81
aes_gcm_enc_128_nt                     @82
aes_gcm_dec_128_nt                     @83
aes_gcm_enc_128_update_nt              @84
aes_gcm_dec_128_update_nt              @85
aes_gcm_enc_256_nt                     @86
aes_gcm_dec_256_nt                     @87
aes_gcm_enc_256_update_nt              @88
aes_gcm_dec_256_update_nt              @89
